# -*- test-case-name: dbxs.test.test_sync_adapter -*-
"""
This adapter can convert any DB-API 2.0 driver to an L{AsyncConnectable}
asynchronous driver, using Twisted's threading infrastructure and returning
Deferreds.

While this adapter does require Twisted be installed, it does I{not}
technically require the Twisted mainloop to be running, if you supply your own
analog to L{twisted.internet.interfaces.IReactorThreads.callFromThread}.
"""

from __future__ import annotations

from dataclasses import dataclass
from queue import Queue
from threading import Thread
from typing import Any, Awaitable, Callable, Optional, Sequence, TypeVar

from twisted._threads import AlreadyQuit, ThreadWorker
from twisted._threads._ithreads import IExclusiveWorker
from twisted.internet.defer import Deferred
from twisted.python.failure import Failure

from ..async_dbapi import (
    AsyncConnectable,
    AsyncConnection,
    AsyncCursor,
    ParamStyle,
)
from ..dbapi import DBAPIColumnDescription, DBAPIConnection, DBAPICursor
from .async_pool import newPool


_T = TypeVar("_T")

F = Callable[[], None]


def _newThread() -> IExclusiveWorker:
    def _startThread(target: Callable[[], None]) -> Thread:
        thread = Thread(target=target, daemon=True)
        thread.start()
        return thread

    return ThreadWorker(_startThread, Queue())


@dataclass
class ExclusiveWorkQueue:
    _worker: Optional[IExclusiveWorker]
    _deliver: Callable[[F], None]

    def worker(self, invalidate: bool = False) -> IExclusiveWorker:
        """
        Assert that the worker should still be present, then return it
        (invalidating it if the flag is passed).
        """
        if invalidate:
            w, self._worker = self._worker, None
        else:
            w = self._worker
        if w is None:
            raise AlreadyQuit("cannot quit twice")
        return w

    def perform(
        self,
        work: Callable[[], _T],
    ) -> Deferred[_T]:
        """
        Perform the given work on the underlying thread, delivering the result
        back to the main thread with L{ExclusiveWorkQueue._deliver}.
        """

        deferred: Deferred[_T] = Deferred()

        def workInThread() -> None:
            try:
                result = work()
            except BaseException:
                f = Failure()
                self._deliver(lambda: deferred.errback(f))
            else:
                self._deliver(lambda: deferred.callback(result))

        self.worker().do(workInThread)

        return deferred

    def quit(self) -> None:
        """
        Allow this thread to stop, and invalidate this L{ExclusiveWorkQueue} by
        removing its C{_worker} attribute.
        """
        self.worker(True).quit()

    def __del__(self) -> None:
        """
        When garbage collected make sure we kill off our underlying thread.
        """
        if self._worker is None:
            return
        # might be nice to emit a ResourceWarning here, since __del__ is not a
        # good way to clean up resources.
        self.quit()


@dataclass
class ThreadedCursorAdapter(AsyncCursor):
    """
    A cursor that can be interacted with asynchronously.
    """

    _cursor: DBAPICursor
    _exclusive: ExclusiveWorkQueue

    async def description(self) -> Optional[Sequence[DBAPIColumnDescription]]:
        result: Optional[
            Sequence[DBAPIColumnDescription]
        ] = await self._exclusive.perform(lambda: self._cursor.description)
        return result

    async def rowcount(self) -> int:
        result: int = await self._exclusive.perform(
            lambda: self._cursor.rowcount
        )
        return result

    async def fetchone(self) -> Optional[Sequence[Any]]:
        result: Optional[Sequence[Any]] = await self._exclusive.perform(
            self._cursor.fetchone
        )
        return result

    # async def fetchmany(
    #     self, size: Optional[int] = None
    # ) -> Sequence[Sequence[Any]]:
    #     a = [size] if size is not None else []
    #     result: Sequence[Sequence[Any]] = await self._exclusive.perform(
    #         lambda: self._cursor.fetchmany(*a)
    #     )
    #     return result

    async def fetchall(self) -> Sequence[Sequence[Any]]:
        result: Sequence[Sequence[Any]] = await self._exclusive.perform(
            self._cursor.fetchall
        )
        return result

    async def execute(
        self,
        operation: str,
        parameters: Sequence[Any] | dict[str, Any] = (),
    ) -> object:
        """
        Execute the given statement.
        """

        def query() -> object:
            return self._cursor.execute(operation, parameters)

        return await self._exclusive.perform(query)

    # async def executemany(
    #     self, __operation: str, __seq_of_parameters: Sequence[Sequence[Any]]
    # ) -> object:
    #     def query() -> object:
    #         return self._cursor.executemany(__operation, __seq_of_parameters)

    #     return await self._exclusive.perform(query)

    async def close(self) -> None:
        """
        Close the underlying cursor.
        """
        await self._exclusive.perform(self._cursor.close)


@dataclass
class _ThreadedConnectionAdapter:
    """
    Asynchronous database connection that binds to a specific thread.
    """

    _connection: Optional[DBAPIConnection]
    _exclusive: ExclusiveWorkQueue
    _paramstyle: ParamStyle

    @property
    def paramstyle(self) -> ParamStyle:
        return self._paramstyle

    def _getConnection(self, invalidate: bool = False) -> DBAPIConnection:
        """
        Get the connection, raising an exception if it's already been
        invalidated.
        """
        c = self._connection
        assert (
            c is not None
        ), "should not be able to get a bad connection via public API"
        if invalidate:
            self._connection = None
        return c

    async def close(self) -> None:
        """
        Close the connection if it hasn't been closed yet.
        """
        connection = self._getConnection(True)
        await self._exclusive.perform(connection.close)
        self._exclusive.quit()

    async def cursor(self) -> ThreadedCursorAdapter:
        """
        Construct a new async cursor.
        """
        c = self._getConnection()
        cur = await self._exclusive.perform(c.cursor)
        return ThreadedCursorAdapter(cur, self._exclusive)

    async def rollback(self) -> None:
        """
        Roll back the current transaction.
        """
        c = self._getConnection()
        await self._exclusive.perform(c.rollback)

    async def commit(self) -> None:
        """
        Roll back the current transaction.
        """
        c = self._getConnection()
        await self._exclusive.perform(c.commit)


def _synthesizeConnector(
    createWorker: Callable[[], IExclusiveWorker],
    deliver: Callable[[Callable[[], None]], None],
    connectCallable: Callable[[], DBAPIConnection],
    paramstyle: ParamStyle,
) -> Callable[[], Awaitable[AsyncConnection]]:
    """
    Creates a callable that creates an AsyncConnection via a threadpool.
    """

    async def connector() -> _ThreadedConnectionAdapter:
        e = ExclusiveWorkQueue(createWorker(), deliver)
        connectedInThread = await e.perform(connectCallable)
        return _ThreadedConnectionAdapter(connectedInThread, e, paramstyle)

    return connector


def adaptSynchronousDriver(
    connectCallable: Callable[[], DBAPIConnection],
    paramstyle: ParamStyle,
    *,
    createWorker: Optional[Callable[[], IExclusiveWorker]] = None,
    callFromThread: Optional[Callable[[F], None]] = None,
    maxIdleConnections: int = 5,
) -> AsyncConnectable:
    """
    Adapt a synchronous DB-API 2.0 driver to be an L{AsyncConnectable} a
    Twisted thread pool.

    @note: If you do not pass your own C{callFromThread} parameter, this
        requires the Twisted reactor to be running in order to process
        responses.
    """
    if callFromThread is None:
        from twisted.internet import reactor

        callFromThread = reactor.callFromThread  # type:ignore[attr-defined]

    if createWorker is None:
        createWorker = _newThread

    return newPool(
        _synthesizeConnector(
            createWorker, callFromThread, connectCallable, paramstyle
        ),
        maxIdleConnections,
    )


__all__ = [
    "adaptSynchronousDriver",
]



"""
Tests for running synchronous DB-API drivers within threads.
"""
from __future__ import annotations

import sqlite3
from contextlib import contextmanager
from dataclasses import dataclass, field
from itertools import count
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from zope.interface import implementer

from twisted._threads import AlreadyQuit
from twisted._threads._ithreads import IExclusiveWorker
from twisted.internet.defer import Deferred
from twisted.trial.unittest import (
    SynchronousTestCase as SyncTestCase,
    TestCase as AsyncTestCase,
)

from .._testing import sqlite3Connector
from .._typing_compat import ParamSpec
from ..adapters.async_pool import _ConnectionPool
from ..adapters.dbapi_twisted import ExclusiveWorkQueue, adaptSynchronousDriver
from ..async_dbapi import AsyncConnectable, InvalidConnection, transaction
from ..dbapi import DBAPIColumnDescription, DBAPIConnection, DBAPICursor


pretendThreadID = 0

_T = TypeVar("_T")
_P = ParamSpec("_P")


def eagerDeferredCoroutine(
    f: Callable[_P, Coroutine[Deferred[Any], object, _T]]
) -> Callable[_P, Deferred[_T]]:
    def inner(*args: _P.args, **kwargs: _P.kwargs) -> Deferred[_T]:
        return Deferred.fromCoroutine(f(*args, **kwargs))

    return inner


@contextmanager
def pretendToBeInThread(newThreadID: int) -> Iterator[None]:
    """
    Pretend to be in the given thread while executing.
    """
    global pretendThreadID
    pretendThreadID = newThreadID
    try:
        yield
    finally:
        pretendThreadID = newThreadID


@dataclass
class FakeDBAPICursor:
    """
    Fake PEP 249 cursor.
    """

    connection: FakeDBAPIConnection
    arraysize: int
    cursorID: int = field(default_factory=count().__next__)

    @property
    def operationsByThread(self) -> List[Tuple[str, int, int]]:
        """
        Delegate to connection.
        """
        return self.connection.operationsByThread

    @property
    def connectionID(self) -> int:
        """
        Delegate to connection.
        """
        return self.connection.connectionID

    @property
    def description(self) -> Optional[Sequence[DBAPIColumnDescription]]:
        # note sqlite actually pads out the response with Nones like this
        self.operationsByThread.append(
            ("description", self.connectionID, pretendThreadID)
        )
        return [("stub", None, None, None, None, None, None)]

    @property
    def rowcount(self) -> int:
        self.operationsByThread.append(
            ("rowcount", self.connectionID, pretendThreadID)
        )
        return 0

    def close(self) -> object:
        self.operationsByThread.append(
            ("close", self.connectionID, pretendThreadID)
        )
        return None

    def execute(
        self,
        operation: str,
        parameters: Union[Sequence[Any], Mapping[str, Any]] = (),
    ) -> object:
        self.operationsByThread.append(
            ("execute", self.connectionID, pretendThreadID)
        )
        return None

    # def executemany(
    #     self, __operation: str, __seq_of_parameters: Sequence[Sequence[Any]]
    # ) -> object:
    #     self.operationsByThread.append(
    #         ("executemany", self.connectionID, pretendThreadID)
    #     )
    #     return None

    def fetchone(self) -> Optional[Sequence[Any]]:
        self.operationsByThread.append(
            ("fetchone", self.connectionID, pretendThreadID)
        )
        return None

    # def fetchmany(self, __size: int = 0) -> Sequence[Sequence[Any]]:
    #     self.operationsByThread.append(
    #         ("fetchmany", self.connectionID, pretendThreadID)
    #     )
    #     return []

    def fetchall(self) -> Sequence[Sequence[Any]]:
        self.operationsByThread.append(
            ("fetchall", self.connectionID, pretendThreadID)
        )
        return []


@dataclass
class FakeDBAPIConnection:
    """
    Fake PEP 249 connection.
    """

    cursors: list[FakeDBAPICursor]
    # (operation, connectionID, threadID)
    operationsByThread: list[tuple[str, int, int]]
    connectionID: int = field(default_factory=count().__next__)

    def close(self) -> None:
        self.operationsByThread.append(
            ("connection.close", self.connectionID, pretendThreadID)
        )
        return None

    def commit(self) -> None:
        self.operationsByThread.append(
            ("commit", self.connectionID, pretendThreadID)
        )
        return None

    def rollback(self) -> Any:
        self.operationsByThread.append(
            ("rollback", self.connectionID, pretendThreadID)
        )

    def cursor(self) -> DBAPICursor:
        self.operationsByThread.append(
            ("cursor", self.connectionID, pretendThreadID)
        )
        cursor = FakeDBAPICursor(self, 3)
        self.cursors.append(cursor)
        return cursor


if TYPE_CHECKING:
    _1: Type[DBAPICursor] = FakeDBAPICursor
    _2: Type[DBAPIConnection] = FakeDBAPIConnection


def realThreadedAdapter(testCase: AsyncTestCase) -> AsyncConnectable:
    """
    Create an AsyncConnectable using real threads and scheduling its
    non-threaded callbacks on the Twisted reactor, suitable for using in a
    real-Deferred-returning integration test.
    """
    memdb = sqlite3Connector()
    scon = memdb()

    scur = scon.cursor()
    scur.execute(
        """
    create table sample (testcol int primary key, testcol2 str)
    """
    )
    scur.execute(
        """
    insert into sample values (1, 'hello'), (2, 'goodbye')
    """
    )
    scon.commit()

    pool = adaptSynchronousDriver(memdb, sqlite3.paramstyle)
    testCase.addCleanup(lambda: Deferred.fromCoroutine(pool.quit()))
    return pool


Thunk = Callable[[], None]


def wrap(t: Thunk, threadID: int) -> Thunk:
    """
    Wrap the given thunk in a fake thread ID
    """

    def wrapped() -> None:
        with pretendToBeInThread(threadID):
            t()

    return wrapped


@implementer(IExclusiveWorker)
@dataclass
class FakeExclusiveWorker:
    queue: List[Thunk]
    threadID: int = field(default_factory=count().__next__)
    quitted: bool = False

    def do(self, work: Callable[[], None]) -> None:
        assert (
            not self.quitted
        ), "we should never schedule work on a quitted worker"
        self.queue.append(wrap(work, self.threadID))

    def quit(self) -> None:
        """
        Exit & clean up.
        """
        self.quitted = True


class SampleError(Exception):
    """
    An error occurred.
    """


class ResourceManagementTests(SyncTestCase):
    """
    Tests to make sure that various thread resources are managed correctly.
    """

    def setUp(self) -> None:
        """
        set up a thread pool that pretends to start threads
        """
        # this queue of work is both "in threads" and "not in threads" work;
        # all "in threads" work is wrapped up in a thing that sets the global
        # L{pretendThreadID}; all "main thread" work has it set to 0.
        self.queue: List[Thunk] = []
        self.cursors: List[FakeDBAPICursor] = []
        self.dbapiops: List[Tuple[str, int, int]] = []
        self.threads: List[FakeExclusiveWorker] = []

        def newWorker() -> IExclusiveWorker:
            w = FakeExclusiveWorker(self.queue)
            self.threads.append(w)
            return w

        def makeConnection() -> DBAPIConnection:
            conn = FakeDBAPIConnection(self.cursors, self.dbapiops)
            self.dbapiops.append(
                ("connect", conn.connectionID, pretendThreadID)
            )
            return conn

        pool = self.pool = adaptSynchronousDriver(
            makeConnection,
            "qmark",
            maxIdleConnections=3,
            createWorker=newWorker,
            callFromThread=self.queue.append,
        )
        assert isinstance(
            pool, _ConnectionPool
        ), "this should be the runtime type"
        self.poolInternals: _ConnectionPool = pool

    def flush(self) -> None:
        """
        Perform all outstanding "threaded" work.
        """
        while self.queue:
            self.queue.pop(0)()

    def test_allOperations(self) -> None:
        """
        All the DB-API operations are wrapped.
        """

        async def dostuff() -> None:
            con = await self.pool.connect()
            cur = await con.cursor()
            self.assertEqual(
                await cur.description(),
                [("stub", None, None, None, None, None, None)],
            )
            self.assertEqual(await cur.rowcount(), 0)
            await cur.execute("test expr", ["some", "params"]), []
            # await cur.executemany(
            #     "lots of operations", [["parameter", "seq"], ["etc", "etc"]]
            # )
            self.assertIs(await cur.fetchone(), None)
            # self.assertEqual(await cur.fetchmany(7), [])
            self.assertEqual(await cur.fetchall(), [])
            await cur.close()
            await con.commit()
            # already committed, so we need a new connection to test rollback
            con2 = await self.pool.connect()
            await con2.rollback()

        d1 = Deferred.fromCoroutine(dostuff())
        self.flush()
        self.successResultOf(d1)
        self.assertOperations(
            [
                "connect",
                "cursor",
                "description",
                "rowcount",
                "execute",
                # "executemany",
                "fetchone",
                # "fetchmany",
                "fetchall",
                "close",
                "close",
                "commit",
                "rollback",
            ],
        )

    def test_connectionClose(self) -> None:
        """
        As opposed to committing or rolling back, closing a connection will
        remove it from the pool entirely.
        """

        async def dostuff() -> None:
            con = await self.pool.connect()
            await con.close()
            with self.assertRaises(InvalidConnection):
                await con.cursor()
            with self.assertRaises(InvalidConnection):
                await con.close()

        d1 = Deferred.fromCoroutine(dostuff())
        self.flush()
        self.successResultOf(d1)
        self.assertOperations(["connect", "connection.close"])
        self.assertEqual(self.poolInternals._idlers, [])
        # The associated thread is also quit.
        self.assertEqual([thread.quitted for thread in self.threads], [True])

    def assertOperations(self, expectedOperations: Sequence[str]) -> None:
        """
        Assert that DB-API would have performed the named operations.
        """
        self.assertEqual(
            expectedOperations, [first for first, _, _ in self.dbapiops]
        )

    def test_inCorrectThread(self) -> None:
        """
        Each connection's operations are executed on a dedicated thread.
        """

        async def dostuff() -> None:
            con = await self.pool.connect()
            cur = await con.cursor()
            await cur.execute("select * from what")
            await con.commit()

        d1 = Deferred.fromCoroutine(dostuff())
        d2 = Deferred.fromCoroutine(dostuff())
        self.assertNoResult(d1)
        self.assertNoResult(d2)
        self.flush()

        self.successResultOf(d1)
        self.successResultOf(d2)

        async def cleanup() -> None:
            await self.pool.quit()

        cleanedup = Deferred.fromCoroutine(cleanup())
        self.flush()

        self.successResultOf(cleanedup)
        self.assertEqual((self.poolInternals._idlers), [])

        threadToConnection: Dict[int, int] = {}
        connectionToThread: Dict[int, int] = {}
        confirmed = 0

        for _, connectionID, threadID in self.dbapiops:
            if threadID in threadToConnection:
                self.assertEqual(threadToConnection[threadID], connectionID)
                self.assertEqual(connectionToThread[connectionID], threadID)
                confirmed += 1
            else:
                threadToConnection[threadID] = connectionID
                connectionToThread[connectionID] = threadID
        # ops = ['connect', 'cursor', 'execute', 'close', 'commit', 'close']
        # expected = (len(ops) * 2) - 2
        expected = 10
        self.assertEqual(confirmed, expected)
        self.assertEqual(len(threadToConnection), 2)

    def test_basicPooling(self) -> None:
        """
        When a pooled connection is committed or rolled back, we will
        invalidate it and won't allocate additional underlying connections.
        """

        async def t1() -> None:
            con = await self.pool.connect()
            await con.commit()
            with self.assertRaises(InvalidConnection):
                await con.cursor()
            con = await self.pool.connect()
            await con.rollback()
            with self.assertRaises(InvalidConnection):
                await con.cursor()

        d = Deferred.fromCoroutine(t1())
        self.flush()
        self.successResultOf(d)
        self.assertEqual(
            len({connectionID for _, connectionID, _ in self.dbapiops}), 1
        )

    def test_tooManyConnections(self) -> None:
        """
        When we exceed the idle-max of the pool, we close connections
        immediately as they are returned.
        """

        async def t1() -> None:
            c1 = await self.pool.connect()
            c2 = await self.pool.connect()
            c3 = await self.pool.connect()
            c4 = await self.pool.connect()
            await c1.commit()
            await c2.commit()
            await c3.commit()
            await c4.commit()

        d = Deferred.fromCoroutine(t1())
        self.flush()
        self.successResultOf(d)
        self.assertEqual(len(self.poolInternals._idlers), 3)
        self.assertOperations(
            [
                *["connect"] * 4,
                *["commit"] * 4,
                "connection.close",
            ]
        )

    def test_transactionContextManager(self) -> None:
        """
        C{with transaction(pool)} results in an async context manager which
        will commit when exited normally and rollback when exited with an
        exception.
        """

        async def t1() -> None:
            # committed
            async with transaction(self.pool) as t:
                await (await t.cursor()).execute("hello world")

            # rolled back
            with self.assertRaises(SampleError):
                async with transaction(self.pool) as t2:
                    await (await t2.cursor()).execute("a")
                    raise SampleError()

        started = Deferred.fromCoroutine(t1())
        self.flush()
        self.successResultOf(started)
        self.assertOperations(
            [
                "connect",
                "cursor",
                "execute",
                "close",
                "commit",
                "cursor",
                "execute",
                "rollback",
            ]
        )

    def test_poolQuit(self) -> None:
        """
        When the pool is shut down, all idlers are closed, and all active
        connections invalidated.
        """

        async def t1() -> None:
            c1 = await self.pool.connect()
            c2 = await self.pool.connect()
            await self.pool.quit()
            with self.assertRaises(InvalidConnection):
                await c1.cursor()
            with self.assertRaises(InvalidConnection):
                await c2.cursor()

        d = Deferred.fromCoroutine(t1())
        self.flush()
        self.successResultOf(d)
        self.assertOperations(
            [
                "connect",
                "connect",
                "rollback",
                "connection.close",
                "rollback",
                "connection.close",
            ]
        )
        self.assertEqual(self.poolInternals._idlers, [])
        self.assertEqual(len(self.threads), 2)
        self.assertEqual(self.threads[0].quitted, True)
        self.assertEqual(self.threads[1].quitted, True)


class InternalSafetyTests(SyncTestCase):
    """
    Tests for internal safety mechanisms; states which I{should} be unreachable
    via the public API but should nonetheless be reported.
    """

    def test_queueQuit(self) -> None:
        """
        L{ExclusiveWorkQueue} should raise L{AlreadyQuit} when interacted with
        after C{quit}.
        """
        stuff: List[Callable[[], None]] = []
        ewc = ExclusiveWorkQueue(FakeExclusiveWorker(stuff), stuff.append)
        ewc.quit()
        with self.assertRaises(AlreadyQuit):
            ewc.quit()
        with self.assertRaises(AlreadyQuit):
            ewc.perform(int)
        self.assertEqual(stuff, [])


class SyncAdapterTests(AsyncTestCase):
    """
    Integration tests for L{adaptSynchronousDriver}.
    """

    @eagerDeferredCoroutine
    async def test_execAndFetch(self) -> None:
        """
        Integration test: can we use an actual DB-API module, with real threads?
        """
        pool = realThreadedAdapter(self)
        con = await pool.connect()
        cur = await con.cursor()

        query = """
        select * from sample order by testcol asc
        """
        await cur.execute(query)
        self.assertEqual(await cur.fetchall(), [(1, "hello"), (2, "goodbye")])
        await cur.execute(
            """
            insert into sample values (3, 'more'), (4, 'even more')
            """
        )
        # await cur.execute(query)
        # self.assertEqual(
        #     await cur.fetchmany(3),
        #     [(1, "hello"), (2, "goodbye"), (3, "more")],
        # )
        # self.assertEqual(await cur.fetchmany(3), [(4, "even more")])

    @eagerDeferredCoroutine
    async def test_errors(self) -> None:
        """
        Integration test: do errors propagate?
        """
        pool = realThreadedAdapter(self)
        con = await pool.connect()
        cur = await con.cursor()
        later = cur.execute("select * from nonexistent")
        with self.assertRaises(sqlite3.OperationalError) as oe:
            await later
        self.assertIn("nonexistent", str(oe.exception))

    @eagerDeferredCoroutine
    async def test_invalidateAfterCommit(self) -> None:
        """
        Connections will be invalidated after they've been committed.
        """
        pool = realThreadedAdapter(self)
        con = await pool.connect()
        await con.commit()
        with self.assertRaises(InvalidConnection):
            await con.cursor()



"""
This module is a description of the abstract types described by U{PEP 249
Database API 2.0 <https://www.python.org/dev/peps/pep-0249/>}.
"""

from __future__ import annotations

from typing import Any, Mapping, Optional, Sequence, Tuple, Union

from ._typing_compat import Protocol


DBAPITypeCode = Optional[Any]


DBAPIColumnDescription = Tuple[
    str,  # name
    DBAPITypeCode,  # type_code
    Optional[int],  # display_size
    Optional[int],  # internal_size
    Optional[int],  # precision
    Optional[int],  # scale
    Union[None, bool, int],  # null_ok
]


class DBAPIConnection(Protocol):
    def close(self) -> object:
        ...

    def commit(self) -> object:
        ...

    def rollback(self) -> Any:
        ...

    def cursor(self) -> DBAPICursor:
        ...


class DBAPICursor(Protocol):
    arraysize: int

    @property
    def description(self) -> Optional[Sequence[DBAPIColumnDescription]]:
        ...

    @property
    def rowcount(self) -> int:
        ...

    def close(self) -> object:
        ...

    def execute(
        self,
        operation: str,
        parameters: Union[Sequence[Any], Mapping[str, Any]] = ...,
    ) -> object:
        ...

    # def executemany(
    #     self, __operation: str, __seq_of_parameters: Sequence[Sequence[Any]]
    # ) -> object:
    #     ...

    def fetchone(self) -> Optional[Sequence[Any]]:
        ...

    # def fetchmany(self, __size: int = ...) -> Sequence[Sequence[Any]]:
    #     ...

    def fetchall(self) -> Sequence[Sequence[Any]]:
        ...



# -*- test-case-name: dbxs.test -*-
from __future__ import annotations

import sqlite3
from dataclasses import dataclass
from typing import Any, Callable, Coroutine, List, TypeVar
from unittest import TestCase
from uuid import uuid4

from twisted._threads._ithreads import IExclusiveWorker
from twisted._threads._memory import createMemoryWorker
from twisted.internet.defer import Deferred
from twisted.python.failure import Failure

from ._typing_compat import Protocol
from .adapters.dbapi_twisted import adaptSynchronousDriver
from .async_dbapi import AsyncConnectable
from .dbapi import DBAPIConnection


def sqlite3Connector() -> Callable[[], DBAPIConnection]:
    """
    Create an in-memory shared-cache SQLite3 database and return a 0-argument
    callable that will connect to that database.
    """
    uri = f"file:{str(uuid4())}?mode=memory&cache=shared"

    held = None

    def connect() -> DBAPIConnection:
        # This callable has to hang on to a connection to the underlying SQLite
        # data structures, otherwise its schema and shared cache disappear as
        # soon as it's garbage collected.  This 'nonlocal' stateemnt adds it to
        # the closure, which keeps the reference after it's created.
        nonlocal held
        return sqlite3.connect(uri, uri=True)

    held = connect()
    return connect


@dataclass
class MemoryPool:
    """
    An in-memory connection pool to an in-memory SQLite database which can be
    controlled a single operation at a time.  Each operation that would
    normally be asynchronously dispatched to a thread can be invoked with the
    L{MemoryPool.pump} and L{MemoryPool.flush} methods.

    @ivar connectable: The L{AsyncConnectable} to be passed to the system under
        test.
    """

    connectable: AsyncConnectable
    _performers: List[Callable[[], bool]]

    def additionalPump(self, f: Callable[[], bool]) -> None:
        """
        Add an additional callable to be called by L{MemoryPool.pump} and
        L{MemoryPool.flush}.  This can be used to interleave other sources of
        in-memory event completion to allow test coroutines to complete, such
        as needing to call U{StubTreq.flush
        <https://treq.readthedocs.io/en/latest/api.html#treq.testing.treq.testing.StubTreq.flush>}
        in a web application.
        """
        self._performers.append(f)

    def pump(self) -> bool:
        """
        Perform one step of pending work.

        @return: True if any work was performed and False if no work was left.
        """
        for performer in self._performers:
            if performer():
                return True
        return False

    def flush(self) -> int:
        """
        Perform all outstanding steps of work.

        @return: a count of the number of steps of work performed.
        """
        steps = 0
        while self.pump():
            steps += 1
        return steps

    @classmethod
    def new(cls) -> MemoryPool:
        """
        Create a synchronous memory connection pool.
        """
        performers = []

        def createWorker() -> IExclusiveWorker:
            worker: IExclusiveWorker
            # note: createMemoryWorker actually returns IWorker, better type
            # annotations may require additional shenanigans
            worker, perform = createMemoryWorker()
            performers.append(perform)
            return worker

        return MemoryPool(
            adaptSynchronousDriver(
                sqlite3Connector(),
                sqlite3.paramstyle,
                createWorker=createWorker,
                callFromThread=lambda f: f(),
                maxIdleConnections=10,
            ),
            performers,
        )


AnyTestCase = TypeVar("AnyTestCase", bound=TestCase)
syncAsyncTest = Callable[
    [AnyTestCase, MemoryPool],
    Coroutine[Any, Any, None],
]
regularTest = Callable[[AnyTestCase], None]


class CompletionTester(Protocol):
    def assertNoResult(self) -> None:
        """
        Raise an exception if a result is present.
        """

    def assertSuccessResult(self) -> None:
        """
        Raise an exception if a result is present.
        """


class ImmediateDriver(Protocol):
    def schedule(
        self,
        failer: Callable[[str], None],
        coroutine: Coroutine[Any, Any, Any],
    ) -> CompletionTester:
        ...


@dataclass
class DeferredCompletionTester:
    failer: Callable[[str], None]
    succeeded: list[object]
    failed: list[Failure]

    def assertNoResult(self) -> None:
        if self.failed:
            self.failed[0].raiseException()
        if self.succeeded:
            self.failer(f"unexpected result {self.succeeded[0]}")

    def assertSuccessResult(self) -> None:
        if self.failed:
            self.failed[0].raiseException()
        if not self.succeeded:
            self.failer("no result")


class ImmediateDeferred:
    @staticmethod
    def schedule(
        failer: Callable[[str], None],
        coroutine: Coroutine[Any, Any, Any],
    ) -> CompletionTester:
        succeeded: list[object] = []
        failed: list[Failure] = []
        deferred = Deferred.fromCoroutine(coroutine)
        deferred.addCallbacks(succeeded.append, failed.append)
        return DeferredCompletionTester(failer, succeeded, failed)


def immediateTest(
    driver: ImmediateDriver = ImmediateDeferred,
) -> Callable[[syncAsyncTest[AnyTestCase]], regularTest[AnyTestCase]]:
    """
    Decorate an C{async def} test that expects a coroutine.
    """

    def decorator(decorated: syncAsyncTest[AnyTestCase]) -> regularTest:
        def regular(self: AnyTestCase) -> None:
            pool = MemoryPool.new()
            d = driver.schedule(self.fail, decorated(self, pool))
            d.assertNoResult()
            while pool.flush():
                pass
            d.assertSuccessResult()

        return regular

    return decorator



# -*- test-case-name: dbxs.test.test_access.AccessTestCase.test_repository -*-
"""
A repository combines a collection of accessors.
"""

from __future__ import annotations

import sys
from contextlib import asynccontextmanager
from inspect import signature
from typing import AsyncContextManager, AsyncIterator, Callable, TypeVar

from ._access import accessor
from .async_dbapi import AsyncConnectable, transaction


T = TypeVar("T")


def repository(
    repositoryType: type[T],
) -> Callable[[AsyncConnectable], AsyncContextManager[T]]:
    """
    A L{repository} combines management of a transaction with management of a
    "repository", which is a collection of L{accessor}s and a contextmanager
    that manages a transaction.  This is easier to show with an example than a
    description::

        class Users(Protocol):
            @query(sql="...", load=one(User))
            def getUserByID(self, id: UserID) -> User: ...

        class Posts(Protocol):
            @query(sql="...", load=many(Post))
            def getPostsFromUser(self, id: UserID) -> AsyncIterator[Posts]: ...

        @dataclass
        class BlogDB:
            users: Users
            posts: Posts

        blogRepository = repository(BlogDB)

        # ...
        async def userAndPosts(pool: AsyncConnectable, id: UserID) -> str:
            async with blogRepository(pool) as blog:
                user = await blog.users.getUserByID(id)
                posts = await blog.posts.getPostsFromUser(posts)
                # transaction commits here
    """

    sig = signature(repositoryType)
    accessors = {}
    for name, parameter in sig.parameters.items():  # pragma: no branch
        annotation = parameter.annotation
        # It would be nicer to do this with signature(..., eval_str=True), but
        # that's not available until we require python>=3.10
        if isinstance(annotation, str):  # pragma: no branch
            annotation = eval(
                annotation, sys.modules[repositoryType.__module__].__dict__
            )
        accessors[name] = accessor(annotation)

    @asynccontextmanager
    async def transactify(acxn: AsyncConnectable) -> AsyncIterator[T]:
        kw = {}
        async with transaction(acxn) as aconn:
            for name in accessors:  # pragma: no branch
                kw[name] = accessors[name](aconn)
            yield repositoryType(**kw)

    return transactify



from unittest import TestCase

from twisted.internet.defer import Deferred

from dbxs.testing import MemoryPool, immediateTest


class TestFailing(TestCase):
    """
    Tests for assertion methods which are normally not invoked in in a
    successful test run.
    """

    def test_immediateFailsToBlock(self) -> None:
        """
        If an @immediateTest fails to block, it fails.
        """

        @immediateTest()
        async def method(self: TestFailing, pool: MemoryPool) -> None:
            ...

        with self.assertRaises(AssertionError):
            method(self)

    def test_immediateBlocksForever(self) -> None:
        """
        If an @immediateTest blocks on something that does not immediately
        return when the memory pool is flushed, it fails.
        """

        @immediateTest()
        async def method(self: TestFailing, pool: MemoryPool) -> None:
            await Deferred()

        with self.assertRaises(AssertionError):
            method(self)

    def test_failsBeforeBlocking(self) -> None:
        """
        If an @immediateTest raises an exception, that exception is re-raised.
        """

        @immediateTest()
        async def method(self: TestFailing, pool: MemoryPool) -> None:
            1 / 0

        with self.assertRaises(ZeroDivisionError):
            method(self)

    def test_failsAfterBlocking(self) -> None:
        """
        If an @immediateTest raises an exception after blocking, that exception
        is re-raised as well.
        """

        @immediateTest()
        async def method(self: TestFailing, pool: MemoryPool) -> None:
            await pool.connectable.connect()
            1 / 0

        with self.assertRaises(ZeroDivisionError):
            method(self)



# -*- test-case-name: dbxs.test.test_access -*-
from __future__ import annotations

from dataclasses import dataclass, field
from inspect import (
    BoundArguments,
    currentframe,
    getsourcefile,
    getsourcelines,
    isawaitable,
    signature,
)
from types import FrameType, TracebackType
from typing import (
    Any,
    AsyncIterable,
    Awaitable,
    Callable,
    Coroutine,
    Dict,
    Iterable,
    List,
    NoReturn,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

from ._typing_compat import ParamSpec, Protocol
from .async_dbapi import AsyncConnection, AsyncCursor


T = TypeVar("T")
P = ParamSpec("P")
A = TypeVar("A", bound=Union[AsyncIterable[object], Awaitable[object]])


class ParamMismatch(Exception):
    """
    The parameters required by the query are different than the parameters
    specified by the function.
    """


class IncorrectResultCount(Exception):
    """
    An assumption about the number of rows from a given query was violated;
    there were either too many or too few.
    """


class NotEnoughResults(IncorrectResultCount):
    """
    There were not enough results for the query to satify L{one}.
    """


class TooManyResults(IncorrectResultCount):
    """
    There were more results for a query than expected; more than one for
    L{one}, or any at all for L{zero}.
    """


class ExtraneousMethods(Exception):
    """
    An access pattern defined extraneous methods.
    """


class WrongRowShape(TypeError):
    """
    The row was the wrong shape for the given callable.
    """


@dataclass
class _ExceptionFixer:
    loader: Callable[..., object]
    definitionLine: int
    decorationLine: int
    decorationFrame: FrameType
    definitionFrame: FrameType

    def reraise(self, row: object, e: Exception) -> NoReturn:
        withDecorationAdded = TracebackType(
            None, self.decorationFrame, 0, self.decorationLine
        )
        withDefinitionAdded = TracebackType(
            withDecorationAdded, self.definitionFrame, 0, self.definitionLine
        )
        raise WrongRowShape(
            f"loader {self.loader.__module__}.{self.loader.__name__}"
            f" could not handle {row}"
        ).with_traceback(withDefinitionAdded) from e

    @classmethod
    def create(cls, loader: Callable[..., T]) -> _ExceptionFixer:
        subFrame = currentframe()
        assert subFrame is not None
        frameworkFrame = subFrame.f_back  # the caller; 'one' or 'many'
        assert frameworkFrame is not None
        realDecorationFrame = frameworkFrame.f_back
        assert realDecorationFrame is not None
        wholeSource, definitionLine = getsourcelines(loader)

        # coverage is tricked by the __code__ modifications below, so we have
        # to explicitly ignore the gap

        def decoratedHere() -> FrameType | None:
            return currentframe()  # pragma: no cover

        def definedHere() -> FrameType | None:
            return currentframe()  # pragma: no cover

        decoratedHere.__code__ = decoratedHere.__code__.replace(
            co_name="<<decorated here>>",
            co_filename=realDecorationFrame.f_code.co_filename,
            co_firstlineno=realDecorationFrame.f_lineno,
        )

        definedSourceFile = getsourcefile(loader)
        definedHere.__code__ = definedHere.__code__.replace(
            co_name="<<defined here>>",
            co_filename=definedSourceFile or "unknown definition",
            co_firstlineno=definitionLine,
        )

        fakeDecorationFrame = decoratedHere()
        definitionFrame = definedHere()
        assert realDecorationFrame is not None
        assert definitionFrame is not None
        assert fakeDecorationFrame is not None

        return cls(
            loader=loader,
            definitionFrame=definitionFrame,
            definitionLine=definitionLine,
            decorationFrame=fakeDecorationFrame,
            decorationLine=realDecorationFrame.f_lineno,
        )


_NR = TypeVar("_NR")


def _makeTranslator(
    fixer: _ExceptionFixer,
    load: Callable[..., T],
    noResults: Callable[[], _NR],
) -> Callable[[object, AsyncCursor], Coroutine[object, object, T | _NR]]:
    async def translator(db: object, cursor: AsyncCursor) -> T | _NR:
        rows = await cursor.fetchall()
        if len(rows) < 1:
            return noResults()
        if len(rows) > 1:
            raise TooManyResults()
        [row] = rows
        try:
            return load(db, *row)
        except TypeError as e:
            fixer.reraise(row, e)

    return translator


def one(
    load: Callable[..., T],
) -> Callable[[object, AsyncCursor], Coroutine[object, object, T]]:
    """
    Fetch a single result with a translator function.
    """
    fixer = _ExceptionFixer.create(load)

    def noResults() -> NoReturn:
        raise NotEnoughResults()

    return _makeTranslator(fixer, load, noResults)


def maybe(
    load: Callable[..., T]
) -> Callable[[object, AsyncCursor], Coroutine[object, object, Optional[T]]]:
    """
    Fetch a single result and pass it to a translator function, but return None
    if it's not found.
    """
    fixer = _ExceptionFixer.create(load)

    def noResults() -> None:
        return None

    return _makeTranslator(fixer, load, noResults)


def many(
    load: Callable[..., T]
) -> Callable[[object, AsyncCursor], AsyncIterable[T]]:
    """
    Fetch multiple results with a function to translate rows.
    """
    fixer = _ExceptionFixer.create(load)

    async def translateMany(
        db: object, cursor: AsyncCursor
    ) -> AsyncIterable[T]:
        while True:
            row = await cursor.fetchone()
            if row is None:
                return
            try:
                yield load(db, *row)
            except TypeError as e:
                fixer.reraise(row, e)

    return translateMany


async def zero(loader: object, cursor: AsyncCursor) -> None:
    """
    Zero record loader.
    """
    result = await cursor.fetchone()
    if result is not None:
        raise TooManyResults("statemnts should not return values")
    return None


METADATA_KEY = "__query_metadata__"


@dataclass
class MaybeAIterable:
    down: Any
    cursor: AsyncCursor = field(init=False)

    def __await__(self) -> Any:
        return self.down.__await__()

    async def __aiter__(self) -> Any:
        try:
            actuallyiter = await self
            async for each in actuallyiter:
                yield each
        finally:
            await self.cursor.close()


@dataclass
class QueryMetadata:
    """
    Metadata defining a certain function on a protocol as a query method.
    """

    sql: str
    load: Callable[[AccessProxy, AsyncCursor], A]
    proxyMethod: Callable[..., Awaitable[object]] = field(init=False)

    def setOn(self, protocolMethod: Any) -> None:
        """
        Attach this QueryMetadata to the given protocol method definition,
        checking its arguments and computing C{proxyMethod} in the process,
        raising L{ParamMismatch} if the expected parameters do not match.
        """
        sig = signature(protocolMethod)
        precomputedSQL: Dict[str, Tuple[str, NameMapMapping]] = {}
        for style, mapFactory in styles.items():
            mapInstance = mapFactory()
            styledSQL = self.sql.format_map(mapInstance)
            precomputedSQL[style] = (styledSQL, mapInstance)

        sampleSQL, sampleInstance = precomputedSQL["qmark"]
        selfExcluded = list(sig.parameters)[1:]
        if set(sampleInstance.names) != set(selfExcluded):
            raise ParamMismatch(
                f"when defining {protocolMethod.__name__}(...), "
                f"SQL placeholders {sampleInstance.names} != "
                f"function params {selfExcluded}"
            )

        def proxyMethod(
            proxySelf: AccessProxy, *args: object, **kw: object
        ) -> Any:
            """
            Implementation of all database-proxy methods on objects returned
            from C{accessor}.

            @note: This should really be two separate methods.  From
                annotations we should know whether to use the aiterable path or
                the awaitable path, we should not need to do the runtime check
                every time.
            """

            maybeai: MaybeAIterable

            async def body() -> Any:
                conn = proxySelf.__query_connection__
                styledSQL, styledMap = precomputedSQL[conn.paramstyle]
                cur = await conn.cursor()
                bound = sig.bind(None, *args, **kw)
                bound.apply_defaults()
                await cur.execute(styledSQL, styledMap.queryArguments(bound))
                maybeAgen: Any = self.load(proxySelf, cur)
                if isawaitable(maybeAgen):
                    # if it's awaitable, then it's not an aiterable.
                    result = await maybeAgen
                    await cur.close()
                    return result
                else:
                    # if it's aiterable, we should be iterating it, not
                    # awaiting it.  MaybeAIterable takes care of the implicit
                    # await of _this_ coroutine.
                    nonlocal maybeai
                    maybeai.cursor = cur
                    return maybeAgen

            maybeai = MaybeAIterable(body())
            return maybeai

        self.proxyMethod = proxyMethod
        setattr(protocolMethod, METADATA_KEY, self)

    @classmethod
    def loadFrom(cls, f: object) -> Optional[QueryMetadata]:
        """
        Load the query metadata for C{f} if it has any.
        """
        self: Optional[QueryMetadata] = getattr(f, METADATA_KEY, None)
        return self

    @classmethod
    def filterProtocolNamespace(
        cls, protocolNamespace: Iterable[Tuple[str, object]]
    ) -> Iterable[Tuple[str, QueryMetadata]]:
        """
        Load all QueryMetadata
        """
        extraneous = []
        for attrname, value in protocolNamespace:
            qm = QueryMetadata.loadFrom(value)
            if qm is None:
                if attrname not in PROTOCOL_IGNORED_ATTRIBUTES:
                    extraneous.append(attrname)
                continue
            yield attrname, qm
        if extraneous:
            raise ExtraneousMethods(
                f"non-query/statement methods defined: {extraneous}"
            )


def query(
    *,
    sql: str,
    load: Callable[[object, AsyncCursor], A],
) -> Callable[[Callable[P, A]], Callable[P, A]]:
    """
    Declare a query method.
    """
    qm = QueryMetadata(sql=sql, load=load)

    def decorator(f: Callable[P, A]) -> Callable[P, A]:
        qm.setOn(f)
        return f

    return decorator


def statement(
    *,
    sql: str,
) -> Callable[
    [Callable[P, Coroutine[Any, Any, None]]],
    Callable[P, Coroutine[Any, Any, None]],
]:
    """
    Declare a query method.
    """
    return query(sql=sql, load=zero)


@dataclass
class DBProxy:
    """
    Database Proxy
    """

    name: str
    transaction: AsyncConnection


@dataclass
class IndexCountingParamstyleMap:
    placeholder: str
    names: List[str] = field(default_factory=list)

    def __getitem__(self, name: str) -> str:
        self.names.append(name)
        return self.placeholder

    def queryArguments(self, bound: BoundArguments) -> Sequence[object]:
        """
        Compute the arguments to the query.
        """
        return [bound.arguments[each] for each in self.names]


class _EmptyProtocol(Protocol):
    """
    Empty protocol for setting a baseline of what attributes to ignore while
    metaprogramming.
    """


PROTOCOL_IGNORED_ATTRIBUTES = set(_EmptyProtocol.__dict__.keys())


class NameMapMapping(Protocol):
    names: List[str]

    def __getitem__(self, __key: str) -> Any:
        ...

    def queryArguments(self, bound: BoundArguments) -> Sequence[object]:
        ...


styles: dict[str, Callable[[], NameMapMapping]] = {
    "qmark": lambda: IndexCountingParamstyleMap("?"),
    "pyformat": lambda: IndexCountingParamstyleMap("%s"),
}


@dataclass
class AccessProxy:
    """
    Superclass of all access proxies.
    """

    __query_connection__: AsyncConnection


def accessor(
    accessPatternProtocol: Callable[[], T]
) -> Callable[[AsyncConnection], T]:
    """
    Create a factory which binds a database transaction in the form of an
    AsyncConnection to a set of declared SQL methods.
    """
    return type(
        f"_{accessPatternProtocol.__name__}_Accessor",
        tuple([AccessProxy]),
        {
            name: metadata.proxyMethod
            for name, metadata in QueryMetadata.filterProtocolNamespace(
                accessPatternProtocol.__dict__.items()
            )
        },
    )



# -*- test-case-name: dbxs.test.test_pg -*-

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Awaitable, Callable, Optional, Sequence, Union

from psycopg import (
    AsyncConnection as PGAsyncConnection,
    AsyncCursor as PGAsyncCursor,
    paramstyle as psycopgParamStyle,
)

from ..async_dbapi import (
    AsyncConnectable,
    AsyncConnection,
    AsyncCursor,
    ParamStyle,
)
from ..dbapi import DBAPIColumnDescription
from .async_pool import newPool


@dataclass
class _PG2DBXSCursor:
    _pgcur: PGAsyncCursor

    async def description(
        self,
    ) -> Optional[Sequence[DBAPIColumnDescription]]:
        subdesc = self._pgcur.description
        if subdesc is None:
            return None
        return [tuple(each) for each in subdesc]  # pragma: no branch

    async def rowcount(self) -> int:
        return self._pgcur.rowcount

    async def fetchone(self) -> Optional[Sequence[Any]]:
        return await self._pgcur.fetchone()

    # async def fetchmany(
    #     self, size: Optional[int] = None
    # ) -> Sequence[Sequence[Any]]:
    #     if size is not None:
    #         return await self._pgcur.fetchmany(size)
    #     else:
    #         return await self._pgcur.fetchmany()

    async def fetchall(self) -> Sequence[Sequence[Any]]:
        return await self._pgcur.fetchall()

    async def execute(
        self,
        operation: str,
        parameters: Union[Sequence[Any], dict[str, Any]] = (),
    ) -> object:
        return await self._pgcur.execute(operation, parameters)

    # async def executemany(
    #     self, __operation: str, __seq_of_parameters: Sequence[Sequence[Any]]
    # ) -> object:
    #     await self._pgcur.executemany(__operation, __seq_of_parameters)
    #     return None

    async def close(self) -> None:
        await self._pgcur.close()


@dataclass
class _PG2DBXSAdapter:
    _pgcon: PGAsyncConnection

    @property
    def paramstyle(self) -> ParamStyle:
        return psycopgParamStyle

    async def cursor(self) -> AsyncCursor:
        return _PG2DBXSCursor(self._pgcon.cursor())

    async def rollback(self) -> None:
        await self._pgcon.rollback()

    async def commit(self) -> None:
        await self._pgcon.commit()

    async def close(self) -> None:
        await self._pgcon.close()


def adaptPostgreSQL(
    connect: Callable[[], Awaitable[PGAsyncConnection]]
) -> AsyncConnectable:
    """
    Adapt a connection created by U{psycopg.AsyncConnection.connect
    <https://www.psycopg.org/psycopg3/docs/api/connections.html#psycopg.AsyncConnection.connect>}
    to an L{AsyncConnection}.
    """

    async def convert() -> AsyncConnection:
        return _PG2DBXSAdapter(await connect())

    return newPool(convert)


__all__ = [
    "adaptPostgreSQL",
]



# -*- test-case-name: dbxs.test.test_pg,dbxs.test.test_mysql-*-
from __future__ import annotations

from abc import abstractmethod
from asyncio.events import get_event_loop
from unittest import TestCase

from ..async_dbapi import AsyncConnectable, transaction


class CommonMeta(type):
    """
    Metaclass to clip out the common base type so that CommonTests itself will
    not be discovered by pyunit or trial.
    """

    def __new__(
        cls, name: str, bases: tuple[type[object], ...], ns: dict[str, object]
    ) -> type[object]:
        bases = (
            ()
            if bases == tuple([TestCase])
            else tuple([CommonTests, TestCase])
        )
        return super().__new__(
            cls,
            name,
            bases,
            ns,
        )


class CommonTests(TestCase, metaclass=CommonMeta):
    @abstractmethod
    def createConnectable(self) -> AsyncConnectable:
        ...

    @abstractmethod
    def numberType(self) -> object:
        ...

    @abstractmethod
    def stringType(self) -> object:
        ...

    @abstractmethod
    def valuesSQL(self) -> str:
        ...

    def test_transaction(self) -> None:
        async def _() -> None:
            cc = self.createConnectable()
            async with transaction(cc) as t:
                cur = await t.cursor()
                await cur.execute(self.valuesSQL())
                desc = await cur.description()
                assert desc is not None
                [
                    (cname, typecode, *otherfirst),
                    (cname2, typecode2, *othersecond),
                ] = desc
                self.assertEqual(cname, "firstcol")
                self.assertEqual(cname2, "secondcol")
                self.assertEqual(self.numberType(), typecode)
                self.assertEqual(self.stringType(), typecode2)
                self.assertEqual((await cur.fetchall()), [(1, "2")])
                await cur.execute("create temporary table foo (bar int)")
                await cur.execute("insert into foo values (1)")
                self.assertEqual(await cur.rowcount(), 1)
                self.assertIsNone(await cur.description())

            with self.assertRaises(ZeroDivisionError):
                async with transaction(cc) as t:
                    cur = await t.cursor()
                    await cur.execute("insert into foo values(2)")
                    1 / 0

            async with transaction(cc) as t:
                cur = await t.cursor()
                await cur.execute("insert into foo values(3)")

            async with transaction(cc) as t:
                cur = await t.cursor()
                await cur.execute("select * from foo order by bar asc")
                rows = await cur.fetchall()
                self.assertEqual(rows, [tuple([1]), tuple([3])])

            await cc.quit()

        get_event_loop().run_until_complete(_())



from __future__ import annotations

from dataclasses import dataclass
from sqlite3 import connect, paramstyle
from typing import AsyncIterable, Protocol


@dataclass
class Quote:
    db: QuoteDB
    id: int
    contents: str


from dbxs import accessor, many, one, query, statement
from dbxs.adapters.dbapi_twisted import adaptSynchronousDriver
from dbxs.async_dbapi import transaction
from dbxs.dbapi import DBAPIConnection


class QuoteDB(Protocol):
    @query(sql="SELECT id, contents FROM quote WHERE id={id}", load=one(Quote))
    async def quoteByID(self, id: int) -> Quote:
        ...

    @query(sql="SELECT id, contents FROM quote", load=many(Quote))
    def allQuotes(self) -> AsyncIterable[Quote]:
        ...

    @statement(sql="INSERT INTO quote (contents) VALUES ({text})")
    async def addQuote(self, text: str) -> None:
        ...


def sqliteWithSchema() -> DBAPIConnection:
    c = connect(":memory:")
    c.execute(
        "CREATE TABLE quote (contents, id INTEGER PRIMARY KEY AUTOINCREMENT)"
    )
    c.commit()
    return c


driver = adaptSynchronousDriver(sqliteWithSchema, paramstyle)
quotes = accessor(QuoteDB)


async def main() -> None:
    async with transaction(driver) as t:
        quotedb: QuoteDB = quotes(t)
        await quotedb.addQuote("hello, world")
        async for quote in quotedb.allQuotes():
            matched = (await quotedb.quoteByID(quote.id)) == quote
            print(f"quote ({quote.id}) {quote.contents!r} {matched}")


if __name__ == "__main__":
    from twisted.internet.defer import Deferred
    from twisted.internet.task import react

    @react
    def run(reactor: object) -> Deferred[None]:
        return Deferred.fromCoroutine(main())



from datetime import datetime
from typing import AsyncIterable, Protocol


class User:
    ...


class Post:
    ...


class PostDB(Protocol):
    async def createUser(self, name: str) -> User:
        ...

    def postsForUser(self, userID: int) -> AsyncIterable[Post]:
        ...

    async def makePostByUser(
        self, created: datetime, content: str, author: int
    ) -> None:
        ...



from __future__ import annotations

from asyncio import get_event_loop
from dataclasses import dataclass
from os import environ
from typing import AsyncIterable
from unittest import skipIf

from twisted.trial.unittest import SynchronousTestCase as TestCase

from dbxs import accessor, many, one, query
from dbxs.async_dbapi import transaction

from .._typing_compat import Protocol
from ..async_dbapi import AsyncConnectable
from .common_adapter_tests import CommonTests


try:
    from psycopg import connect
except ImportError:
    cantFindPG = "psycopg not installed"
else:
    from psycopg import NUMBER, STRING, AsyncConnection

    from ..adapters.async_psycopg import adaptPostgreSQL

    try:
        with connect() as con:
            with con.cursor() as cur:
                cur.execute("select true")
                assert cur.fetchall() == [tuple([True])]
                cantFindPG = ""
    except Exception as e:  # pragma: no cover
        cantFindPG = f"could not connect: {e} ({environ.get('PGPORT')})"


@dataclass
class SimpleRow:
    db: PGInternalsAccess
    name: str
    value: int


class PGInternalsAccess(Protocol):
    @query(sql="select version()", load=one(lambda db, x: str(x)))
    async def getPostgresVersion(self) -> str:
        ...

    @query(
        sql="""select * from
        (values ({firstName}, {firstValue}), ({secondName}, {secondValue}))
        as simple_rows(name, value)
        """,
        load=many(SimpleRow),
    )
    def cannedValues(
        self,
        firstName: str,
        firstValue: int,
        secondName: str,
        secondValue: int,
    ) -> AsyncIterable[SimpleRow]:
        ...


pgia = accessor(PGInternalsAccess)


class AccessTestCase(TestCase):
    if cantFindPG:
        skip = cantFindPG

    def test_connect(self) -> None:
        """
        Testing self-test to ensure we can still connect to the database once
        the tests are running.
        """
        with connect() as con:
            with con.cursor() as cur:
                cur.execute("select true")
                self.assertEqual(cur.fetchall(), [tuple([True])])

    def test_basicAsyncConnection(self) -> None:
        async def _() -> None:
            self.assertIn(
                "PostgreSQL",
                await pgia(
                    await adaptPostgreSQL(AsyncConnection.connect).connect()
                ).getPostgresVersion(),
            )

        get_event_loop().run_until_complete(_())

    def test_valueConversions(self) -> None:
        async def _() -> None:
            everything = []
            adapted = await adaptPostgreSQL(AsyncConnection.connect).connect()
            access = pgia(adapted)
            values = access.cannedValues("hello", 1, "second", 2)
            async for row in values:
                everything.append((row.name, row.value))
            self.assertEqual([("hello", 1), ("second", 2)], everything)

        get_event_loop().run_until_complete(_())

    def test_transaction(self) -> None:
        async def _() -> None:
            async with transaction(
                adaptPostgreSQL(AsyncConnection.connect)
            ) as t:
                cur = await t.cursor()
                await cur.execute(
                    "select * from (values (1, '2')) "
                    "as named(firstcol, secondcol)"
                )
                desc = await cur.description()
                assert desc is not None
                [
                    (cname, typecode, *otherfirst),
                    (cname2, typecode2, *othersecond),
                ] = desc
                self.assertEqual(cname, "firstcol")
                self.assertEqual(cname2, "secondcol")
                self.assertEqual(NUMBER, typecode)
                self.assertEqual(STRING, typecode2)

        get_event_loop().run_until_complete(_())


@skipIf(cantFindPG, cantFindPG)
class PGTests(CommonTests):
    def createConnectable(self) -> AsyncConnectable:
        return adaptPostgreSQL(AsyncConnection.connect)

    def numberType(self) -> object:
        return NUMBER

    def stringType(self) -> object:
        return STRING

    def valuesSQL(self) -> str:
        return "select * from (values (1, '2')) as named(firstcol, secondcol)"


# print(PGTests.skip)



"""
DBXS (“D.B. Access”) is an SQL database access layer for Python.
"""

from ._access import (
    ExtraneousMethods,
    IncorrectResultCount,
    NotEnoughResults,
    ParamMismatch,
    TooManyResults,
    accessor,
    many,
    maybe,
    one,
    query,
    statement,
)
from ._repository import repository


__version__ = "0.1.0"


__all__ = [
    "one",
    "many",
    "maybe",
    "accessor",
    "repository",
    "statement",
    "query",
    "ParamMismatch",
    "TooManyResults",
    "NotEnoughResults",
    "IncorrectResultCount",
    "ExtraneousMethods",
]



"""
Since we support a range of Python and Mypy versions where certain features are
available across L{typing} and L{typing_extensions}, we put those aliases here
to avoid repeating conditional import logic.
"""
import sys


if sys.version_info > (3, 10):
    from typing import Concatenate, ParamSpec, Protocol
else:
    # PyPy 3.9 seems to have a bonus runtime check for Protocol's generic
    # arguments all being TypeVars, so lie to it about ParamSpec.
    from typing import TYPE_CHECKING

    from typing_extensions import Concatenate, Protocol

    if TYPE_CHECKING:
        from typing_extensions import ParamSpec
    else:
        from platform import python_implementation

        if python_implementation() == "PyPy":
            from typing import TypeVar as ParamSpec
        else:
            from typing_extensions import ParamSpec


__all__ = [
    "Protocol",
    "ParamSpec",
    "Concatenate",
]



# -*- test-case-name: dbxs.test.test_mysql -*-
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Awaitable, Callable, Optional, Sequence, Union

from mysql.connector import paramstyle as mysqlParamStyle
from mysql.connector.aio.abstracts import (
    MySQLConnectionAbstract,
    MySQLCursorAbstract,
)

from ..async_dbapi import (
    AsyncConnectable,
    AsyncConnection,
    AsyncCursor,
    ParamStyle,
)
from ..dbapi import DBAPIColumnDescription
from .async_pool import newPool


@dataclass
class _MYSQL2DBXSCursor:
    _mysqlcur: MySQLCursorAbstract

    async def description(
        self,
    ) -> Optional[Sequence[DBAPIColumnDescription]]:
        real = self._mysqlcur.description
        if real is None:
            return None
        return [desc[:7] for desc in real]  # pragma: no branch

    async def rowcount(self) -> int:
        return self._mysqlcur.rowcount

    async def fetchone(self) -> Optional[Sequence[Any]]:
        return await self._mysqlcur.fetchone()

    # async def fetchmany(
    #     self, size: Optional[int] = None
    # ) -> Sequence[Sequence[Any]]:
    #     if size is not None:
    #         return await self._mysqlcur.fetchmany(size)
    #     else:
    #         return await self._mysqlcur.fetchmany()

    async def fetchall(self) -> Sequence[Sequence[Any]]:
        return await self._mysqlcur.fetchall()

    async def execute(
        self,
        operation: str,
        parameters: Union[Sequence[Any], dict[str, Any]] = (),
    ) -> object:
        await self._mysqlcur.execute(operation, parameters)
        return None

    # async def executemany(
    #     self, __operation: str, __seq_of_parameters: Sequence[Sequence[Any]]
    # ) -> object:
    #     await self._mysqlcur.executemany(__operation, __seq_of_parameters)
    #     return None

    async def close(self) -> None:
        await self._mysqlcur.close()


@dataclass
class _MYSQL2DBXSAdapter:
    _mysqlcon: MySQLConnectionAbstract

    @property
    def paramstyle(self) -> ParamStyle:
        return mysqlParamStyle

    async def cursor(self) -> AsyncCursor:
        return _MYSQL2DBXSCursor(await self._mysqlcon.cursor())

    async def rollback(self) -> None:
        await self._mysqlcon.rollback()

    async def commit(self) -> None:
        await self._mysqlcon.commit()

    async def close(self) -> None:
        await self._mysqlcon.close()


def adaptMySQL(
    connection: Callable[[], Awaitable[MySQLConnectionAbstract]]
) -> AsyncConnectable:
    """
    Adapt a connection created by U{mysql.connector.aio.connect
    <https://dev.mysql.com/doc/connector-python/en/connector-python-asyncio.html>}
    to an L{AsyncConnection}.
    """

    async def convert() -> AsyncConnection:
        return _MYSQL2DBXSAdapter(await connection())

    return newPool(convert)


__all__ = [
    "adaptMySQL",
]



"""
Testing support for L{dbxs}.

L{MemoryPool} creates a synchronous, in-memory SQLite database that can be used
for testing anything that needs an
L{dbxs.dbapi_async.AsyncConnectable}.
"""

from ._testing import MemoryPool, immediateTest


__all__ = [
    "MemoryPool",
    "immediateTest",
]



from __future__ import annotations

import traceback
from dataclasses import dataclass
from typing import AsyncIterable, Optional
from unittest import TestCase

from .. import (
    ExtraneousMethods,
    NotEnoughResults,
    ParamMismatch,
    TooManyResults,
    accessor,
    many,
    maybe,
    one,
    query,
    repository,
    statement,
)
from .._typing_compat import Protocol
from ..async_dbapi import AsyncConnection, transaction
from ..testing import MemoryPool, immediateTest


# Trying to stick to the public API for what we're testing; no underscores here.


@dataclass
class Foo:
    db: FooAccessPattern
    bar: int
    baz: int


def oops(  # point at this definition(one)
    db: FooAccessPattern,
    bar: int,
    baz: int,
    extra: str,
) -> str:
    return extra  # pragma: no cover


# duplicate definition comment on different lines below because
# inspect.getsourcelines changed behavior from 3.8 to 3.9


@dataclass  # point at this definition(many)
class Oops2:  # point at this definition(many)
    db: FooAccessPattern
    bar: int
    baz: int
    extra: str


class FooAccessPattern(Protocol):
    @query(sql="select bar, baz from foo where bar = {bar}", load=one(Foo))
    async def getFoo(self, bar: int) -> Foo:
        ...

    @query(
        sql="select bar, baz from foo order by bar asc",
        load=many(Foo),
    )
    def allFoos(self) -> AsyncIterable[Foo]:
        ...

    @query(sql="select bar, baz from foo where bar = {bar}", load=maybe(Foo))
    async def maybeFoo(self, bar: int) -> Optional[Foo]:
        ...

    @query(sql="select bar, baz from foo where baz = {baz}", load=one(Foo))
    async def oneFooByBaz(self, baz: int) -> Foo:
        ...

    @query(
        sql="select bar, baz from foo where baz = {baz}",
        load=one(oops),  # point at this decoration(one)
    )
    async def wrongArityOne(self, baz: int) -> str:
        ...

    @query(
        sql="select bar, baz from foo",
        load=many(Oops2),  # point at this decoration(many)
    )
    def wrongArityMany(self) -> AsyncIterable[Oops2]:
        ...

    @query(sql="select bar, baz from foo where baz = {baz}", load=maybe(Foo))
    async def maybeFooByBaz(self, baz: int) -> Optional[Foo]:
        ...

    @statement(sql="insert into foo (baz) values ({baz})")
    async def newFoo(self, baz: int) -> None:
        """
        Create a new C{Foo}
        """

    @statement(sql="select * from foo")
    async def oopsQueryNotStatement(self) -> None:
        """
        Oops, it's a query, not a statement, it returns values.
        """

    @query(sql="select {value}", load=one(lambda db, x: str(x)))
    async def echoValue(self, value: int = 3) -> str:
        """
        Echo the given value back, with a default provided.
        """

    @query(
        sql="insert into foo (baz) values ({baz}) returning bar, baz",
        load=one(Foo),
    )
    async def newReturnFoo(self, baz: int) -> Foo:
        """
        Create a new C{Foo} and return it.
        """


class OtherAccessPattern(Protocol):
    @query(sql="select {value} + 1", load=one(lambda db, x: x))
    async def addOneTo(self, value: int) -> int:
        ...


accessFoo = accessor(FooAccessPattern)


@dataclass
class ExampleRepository:
    foo: FooAccessPattern
    other: OtherAccessPattern


exampleRepo = repository(ExampleRepository)


async def schemaAndData(c: AsyncConnection) -> None:
    """
    Create the schema for 'foo' and insert some sample data.
    """
    cur = await c.cursor()
    for stmt in """
        create table foo (bar integer primary key autoincrement, baz int);
        insert into foo values (1, 3);
        insert into foo values (2, 4);
        """.split(
        ";"
    ):
        await cur.execute(stmt)


class AccessTestCase(TestCase):
    """
    Tests for L{accessor} and its associated functions
    """

    @immediateTest()
    async def test_happyPath(self, pool: MemoryPool) -> None:
        """
        Declaring a protocol with a query and executing it
        """
        async with transaction(pool.connectable) as c:
            await schemaAndData(c)
            db = accessFoo(c)
            result = await db.getFoo(1)
            result2 = await db.maybeFoo(1)
            result3 = [  # pragma: no branch
                each async for each in db.allFoos()
            ]
        self.assertEqual(result, Foo(db, 1, 3))
        self.assertEqual(result, result2)
        self.assertEqual(result3, [Foo(db, 1, 3), Foo(db, 2, 4)])

    @immediateTest()
    async def test_defaultParamValue(self, pool: MemoryPool) -> None:
        """
        Default parameters specified by the access Protocol are incorporated
        into the query placeholders.
        """
        async with transaction(pool.connectable) as c:
            await schemaAndData(c)
            db = accessFoo(c)
            result = await db.echoValue(7)
            self.assertEqual(result, "7")
            result = await db.echoValue()
            self.assertEqual(result, "3")

    @immediateTest()
    async def test_wrongResultArity(self, pool: MemoryPool) -> None:
        """
        If the signature of the callable provided to C{query(load=one(...))} or
        C{query(load=many(...))} does not match with the number of arguments
        returned by the database for a row in a particular query, the error
        will explain well enough to debug.
        """
        async with transaction(pool.connectable) as c:
            await schemaAndData(c)
            db = accessFoo(c)
            try:
                await db.wrongArityOne(3)
            except TypeError:
                tbf1 = traceback.format_exc()
            try:
                [  # pragma: no branch
                    each async for each in db.wrongArityMany()
                ]
            except TypeError:
                tbf2 = traceback.format_exc()
            # print(tbf1)
            # print(tbf2)
            self.assertIn("point at this definition(one)", tbf1)
            self.assertIn("point at this decoration(one)", tbf1)
            self.assertIn("point at this definition(many)", tbf2)
            self.assertIn("point at this decoration(many)", tbf2)

    def test_argumentExhaustiveness(self) -> None:
        """
        If a query does not use all of its arguments, or the function does not
        specify all the arguments that a function uses, it will raise an
        exception during definition.
        """
        with self.assertRaises(ParamMismatch) as pm:

            class MissingBar(Protocol):
                @statement(sql="fake sql {bar}")
                async def someUnused(self) -> None:
                    ...

        self.assertIn("bar", str(pm.exception))
        self.assertIn("someUnused", str(pm.exception))
        with self.assertRaises(ParamMismatch):

            class DoesntUseBar(Protocol):
                @statement(sql="fake sql")
                async def someMissing(self, bar: str) -> None:
                    ...

    @immediateTest()
    async def test_tooManyResults(self, pool: MemoryPool) -> None:
        """
        If there are too many results for a L{one} query, then a
        L{TooManyResults} exception is raised.
        """
        async with transaction(pool.connectable) as c:
            await schemaAndData(c)
            cur = await c.cursor()
            await cur.execute("insert into foo (baz) values (3)")
            await cur.execute("insert into foo (baz) values (3)")
            db = accessFoo(c)
            with self.assertRaises(TooManyResults):
                await db.oneFooByBaz(3)
            with self.assertRaises(TooManyResults):
                await db.maybeFooByBaz(3)

    def test_brokenProtocol(self) -> None:
        """
        Using L{accessor} on a protocol with unrelated methods raises a .
        """

        class NonAccessPatternProtocol(Protocol):
            def randomNonQueryMethod(self) -> None:
                ...

        with self.assertRaises(ExtraneousMethods) as em:
            accessor(NonAccessPatternProtocol)
        self.assertIn("randomNonQueryMethod", str(em.exception))

    @immediateTest()
    async def test_notEnoughResults(self, pool: MemoryPool) -> None:
        """
        If there are too many results for a L{one} query, then a
        L{NotEnoughResults} exception is raised.
        """
        async with transaction(pool.connectable) as c:
            cur = await c.cursor()
            await schemaAndData(c)
            await cur.execute("delete from foo")
            db = accessFoo(c)
            with self.assertRaises(NotEnoughResults):
                await db.getFoo(1)
            self.assertIs(await db.maybeFoo(1), None)

    @immediateTest()
    async def test_insertStatementWithReturn(self, pool: MemoryPool) -> None:
        """
        DML statements can use RETURNING to return values.
        """
        async with transaction(pool.connectable) as c:
            await schemaAndData(c)
            db = accessFoo(c)
            self.assertEqual(await db.newReturnFoo(100), Foo(db, 3, 100))

    @immediateTest()
    async def test_statementHasNoResult(self, pool: MemoryPool) -> None:
        """
        The L{statement} decorator gives a result.
        """
        async with transaction(pool.connectable) as c:
            await schemaAndData(c)
            db = accessFoo(c)
            nothing = await db.newFoo(7)  # type:ignore[func-returns-value]
            self.assertIs(nothing, None)

    @immediateTest()
    async def test_statementWithResultIsError(self, pool: MemoryPool) -> None:
        """
        The L{statement} decorator gives a result.
        """
        async with transaction(pool.connectable) as c:
            await schemaAndData(c)
            db = accessFoo(c)
            with self.assertRaises(TooManyResults) as tmr:
                await db.oopsQueryNotStatement()
            self.assertIn("should not return", str(tmr.exception))

    @immediateTest()
    async def test_repository(self, pool: MemoryPool) -> None:
        """
        Test constructing a repository.
        """
        async with exampleRepo(pool.connectable) as repo:
            self.assertEqual(await repo.foo.echoValue(7), "7")
            self.assertEqual(await repo.other.addOneTo(3), 4)



"""
This is a collection of abstract types that are I{like} the PEP 249 types in
L{dbxs.dbapi}, but with C{await} put in the relevant places to make them
asynchronous.
"""

from __future__ import annotations

from contextlib import asynccontextmanager
from typing import Any, AsyncIterator, Optional, Sequence, TypeVar, Union

from ._typing_compat import Protocol
from .dbapi import DBAPIColumnDescription


class InvalidConnection(Exception):
    """
    The connection has already been closed, or the transaction has already been
    committed.
    """


ParamStyle = str

# Sadly, db-api modules do not restrict themselves in this way, so we can't
# specify the ParamStyle type more precisely, like so:

# ParamStyle = Literal['qmark', 'numeric', 'named', 'format', 'pyformat']

T = TypeVar("T")


class AsyncCursor(Protocol):
    """
    Asynchronous Cursor Object.
    """

    async def description(
        self,
    ) -> Optional[Sequence[DBAPIColumnDescription]]:
        ...

    async def rowcount(self) -> int:
        ...

    async def fetchone(self) -> Optional[Sequence[Any]]:
        ...

    # async def fetchmany(
    #     self, size: Optional[int] = None
    # ) -> Sequence[Sequence[Any]]:
    #     ...

    async def fetchall(self) -> Sequence[Sequence[Any]]:
        ...

    async def execute(
        self,
        operation: str,
        parameters: Union[Sequence[Any], dict[str, Any]] = (),
    ) -> object:
        ...

    # async def executemany(
    #     self, __operation: str, __seq_of_parameters: Sequence[Sequence[Any]]
    # ) -> object:
    #     ...

    async def close(self) -> None:
        ...


class AsyncConnection(Protocol):
    """
    Asynchronous version of a DB-API connection.
    """

    @property
    def paramstyle(self) -> ParamStyle:
        ...

    async def cursor(self) -> AsyncCursor:
        ...

    async def rollback(self) -> None:
        ...

    async def commit(self) -> None:
        ...

    async def close(self) -> None:
        ...


class AsyncConnectable(Protocol):
    """
    An L{AsyncConnectable} can establish and pool L{AsyncConnection} objects.
    """

    async def connect(self) -> AsyncConnection:
        ...

    async def quit(self) -> None:
        ...


@asynccontextmanager
async def transaction(
    connectable: AsyncConnectable,
) -> AsyncIterator[AsyncConnection]:
    """
    Connect to a given connection in a context manager.
    """
    conn = await connectable.connect()
    try:
        yield conn
    except BaseException:
        await conn.rollback()
        raise
    else:
        await conn.commit()



from __future__ import annotations

from asyncio import get_event_loop
from dataclasses import dataclass
from os import environ
from typing import AsyncIterable
from unittest import skipIf

from twisted.trial.unittest import SynchronousTestCase as TestCase

from dbxs import accessor, many, one, query

from .._typing_compat import Protocol
from ..async_dbapi import AsyncConnectable
from .common_adapter_tests import CommonTests


try:
    from mysql.connector import connect
except ImportError:
    cantFindMySQL = "mysql-connector-python not installed"
else:
    from mysql.connector import NUMBER, STRING
    from mysql.connector.aio import connect as connectAsync
    from mysql.connector.aio.abstracts import MySQLConnectionAbstract

    from ..adapters.async_mysql import adaptMySQL

    try:
        with connect(
            user=environ["MYSQL_USER"], password=environ["MYSQL_PWD"]
        ) as con:
            with con.cursor() as cur:
                cur.execute("select true")
                assert cur.fetchall() == [tuple([True])]
                cantFindMySQL = ""
    except Exception as e:  # pragma: no cover
        cantFindMySQL = f"could not connect: {e} ({environ.get('MYSQL_USER')})"


async def configuredConnectAsync() -> MySQLConnectionAbstract:
    connected = await connectAsync(
        user=environ["MYSQL_USER"], password=environ["MYSQL_PWD"]
    )
    cur = await connected.cursor()
    await cur.execute("create database if not exists dbxs_test_suite")
    await cur.execute("use dbxs_test_suite")
    await cur.close()
    return connected


@dataclass
class SimpleRow:
    db: MySQLInternalsAccess
    name: str
    value: int


def justStringify(
    db: MySQLInternalsAccess, varName: str, varValue: str
) -> str:
    return str(varValue)


class MySQLInternalsAccess(Protocol):
    @query(
        sql="show variables like 'version_comment'", load=one(justStringify)
    )
    async def getMysqlVersion(self) -> str:
        ...

    @query(
        sql="""select * from
        (values ROW({firstName}, {firstValue}),
         ROW({secondName}, {secondValue}))
        as simple_rows(name, value)
        """,
        load=many(SimpleRow),
    )
    def cannedValues(
        self,
        firstName: str,
        firstValue: int,
        secondName: str,
        secondValue: int,
    ) -> AsyncIterable[SimpleRow]:
        ...


pgia = accessor(MySQLInternalsAccess)


class AccessTestCase(TestCase):
    if cantFindMySQL:
        skip = cantFindMySQL

    def test_basicAsyncConnection(self) -> None:
        async def _() -> None:
            self.assertIn(
                "MySQL",
                await pgia(
                    await adaptMySQL(configuredConnectAsync).connect()
                ).getMysqlVersion(),
            )

        get_event_loop().run_until_complete(_())

    def test_valueConversions(self) -> None:
        async def _() -> None:
            everything = []
            async for row in pgia(
                await adaptMySQL(configuredConnectAsync).connect()
            ).cannedValues("hello", 1, "second", 2):
                everything.append((row.name, row.value))
            self.assertEqual([("hello", 1), ("second", 2)], everything)

        get_event_loop().run_until_complete(_())


@skipIf(cantFindMySQL, cantFindMySQL)
class MySQLTests(CommonTests):
    def createConnectable(self) -> AsyncConnectable:
        return adaptMySQL(configuredConnectAsync)

    def numberType(self) -> object:
        return NUMBER

    def stringType(self) -> object:
        return STRING

    def valuesSQL(self) -> str:
        return (
            "select * from (values row(1, '2')) as named(firstcol, secondcol)"
        )



from __future__ import annotations

import sqlite3
from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING, AsyncIterable, Protocol

from twisted.internet.defer import Deferred
from twisted.internet.interfaces import IReactorCore
from twisted.python.failure import Failure

from dbxs import many, one, query, repository, statement
from dbxs.adapters.dbapi_twisted import adaptSynchronousDriver
from dbxs.async_dbapi import transaction


schema = """
CREATE TABLE IF NOT EXISTS user (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    name TEXT NOT NULL
);
CREATE TABLE IF NOT EXISTS post (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    created TIMESTAMP NOT NULL,
    content TEXT NOT NULL,
    author INTEGER NOT NULL,
    FOREIGN KEY(author)
        REFERENCES user(id)
        ON DELETE CASCADE
);
"""


def newConnection() -> sqlite3.dbapi2.Connection:
    result = sqlite3.connect("user-posts.sqlite")
    return result


asyncDriver = adaptSynchronousDriver(
    (lambda: newConnection()), sqlite3.paramstyle
)


# user attributes
@dataclass
class User:
    postDB: PostDB
    id: int
    name: str
    # end user attributes

    async def post(self, text: str) -> None:
        return await self.postDB.makePostByUser(datetime.now(), text, self.id)

    def posts(self) -> AsyncIterable[Post]:
        return self.postDB.postsForUser(self.id)


@dataclass
class Post:
    postDB: PostDB
    created: datetime
    content: str
    id: int
    what: object


class PostDB(Protocol):
    @query(
        sql="""
        insert into user(name)
        values({name})
        returning id, name
        """,
        load=one(User),
    )
    async def createUser(self, name: str) -> User:
        ...

    @query(
        sql="""
        select created, content, author, id
        from post
        where author = {userID}
        """,
        load=many(Post),
    )
    def postsForUser(self, userID: int) -> AsyncIterable[Post]:
        ...

    @statement(
        sql="""
        insert into post( created,   content,   author)
        values          ({created}, {content}, {author})
        """
    )
    async def makePostByUser(
        self, created: datetime, content: str, author: int
    ) -> None:
        ...


@dataclass
class BlogRepo:
    posts: PostDB


blog = repository(BlogRepo)


async def main() -> None:
    async with transaction(asyncDriver) as c:
        cur = await c.cursor()
        for expr in schema.split(";"):
            await cur.execute(expr)

    async with blog(asyncDriver) as db:
        b = await db.posts.createUser("bob")
        await b.post("a post")
        await b.post("another post")
        post: Post
        async for post in b.posts():
            print(post.created, repr(post.content))


if __name__ == "__main__":
    reactor: IReactorCore
    if not TYPE_CHECKING:
        from twisted.internet import reactor

    def reportAndStop(f: Failure | None) -> None:
        reactor.stop()
        if f is not None:
            print(f)
        else:
            return f
        print("STOP")

    reactor.callWhenRunning(
        lambda: (Deferred.fromCoroutine(main()).addBoth(reportAndStop))
    )
    reactor.run()



"""
Connection pooling for L{AsyncConnection}.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Awaitable, Callable, List, Optional

from ..async_dbapi import (
    AsyncConnectable,
    AsyncConnection,
    AsyncCursor,
    InvalidConnection,
)


@dataclass(eq=False)
class _PooledConnectionGuard:
    """
    Pooled connection adapter that re-adds itself back to the pool upon commit
    or rollback, and guards against direct access to the underlying connection,
    to avoid changing its state behind the back of the L{_ConnectionPool}.
    """

    _adapter: Optional[AsyncConnection]
    _pool: _ConnectionPool
    _cursors: List[AsyncCursor]

    def _original(self, invalidate: bool) -> AsyncConnection:
        """
        Check for validity, return the underlying connection, and then
        optionally invalidate this adapter.
        """
        a = self._adapter
        if a is None:
            raise InvalidConnection("The connection has already been closed.")
        if invalidate:
            self._adapter = None
        return a

    @property
    def paramstyle(self) -> str:
        return self._original(False).paramstyle

    async def cursor(self) -> AsyncCursor:
        it = await self._original(False).cursor()
        self._cursors.append(it)
        return it

    async def rollback(self) -> None:
        """
        Roll back the transaction, returning the connection to the pool.
        """
        a = self._original(True)
        try:
            await a.rollback()
        finally:
            await self._pool._checkin(self, a)

    async def _closeCursors(self) -> None:
        for cursor in self._cursors:
            await cursor.close()

    async def commit(self) -> None:
        """
        Commit the transaction, returning the connection to the pool.
        """
        await self._closeCursors()
        a = self._original(True)
        try:
            await a.commit()
        finally:
            await self._pool._checkin(self, a)

    async def close(self) -> None:
        """
        Close the underlying connection, removing it from the pool.
        """
        await self._closeCursors()
        await self._original(True).close()


@dataclass(eq=False)
class _ConnectionPool:
    """
    Database engine and connection pool.
    """

    _newConnection: Callable[[], Awaitable[AsyncConnection]]
    _idleMax: int
    _idlers: List[AsyncConnection] = field(default_factory=list)
    _active: List[_PooledConnectionGuard] = field(default_factory=list)

    async def connect(self) -> AsyncConnection:
        """
        Checkout a new connection from the pool, connecting to the database and
        opening a thread first if necessary.
        """
        txn = _PooledConnectionGuard(
            (
                self._idlers.pop()
                if self._idlers
                else await self._newConnection()
            ),
            self,
            [],
        )
        self._active.append(txn)
        return txn

    async def _checkin(
        self,
        txn: _PooledConnectionGuard,
        connection: AsyncConnection,
    ) -> None:
        """
        Check a connection back in to the pool, closing and discarding it.
        """
        self._active.remove(txn)
        if len(self._idlers) < self._idleMax:
            self._idlers.append(connection)
        else:
            await connection.close()

    async def quit(self) -> None:
        """
        Close all outstanding connections and shut down the underlying
        threadpool.
        """
        self._idleMax = 0
        while self._active:
            await self._active[0].rollback()

        while self._idlers:
            await self._idlers.pop().close()


def newPool(
    newConnection: Callable[[], Awaitable[AsyncConnection]],
    maxIdleConnections: int = 5,
) -> AsyncConnectable:
    """
    Create a new connection pool that wraps up a callable that can create an
    L{AsyncConnection}.
    """
    return _ConnectionPool(newConnection, maxIdleConnections)


__all__ = ["newPool"]
